import itertools
import sys
import time
from rich.markdown import Markdown
from rich.console import Console
from rich.text import Text
from langchain_openai import ChatOpenAI
from prompt_toolkit.formatted_text import HTML

import re
import subprocess
import threading

from nucleus.logger import log

console = Console()

class MessagePrinter:
    def __init__(self):
        self.console = Console()
        self.style="#fafcfb"

    def system_message(self, message):
        """
        Prints a system message with specific styling.
        """
        system_text = Text(message, style="yellow")
        self.console.print(system_text)

    def user_message(self, message):
        """
        Prints a user message with Markdown rendering.
        """
        user_text = Text(f"{message}")
        self.console.print(user_text, style=self.style)

    def assistant_message(self, message, type=None):
        """
        Prints an assistant message with optional Markdown rendering.
        """
        if type=='command':
            print("\n"+ message)
        else:
            # pprint(message)
            assistant_text = Markdown(f"<br>{message}")
            self.console.print(assistant_text)

def command_provider(message , model_name):
    """
    """

    llm = None

    if model_name == 'openai':
        llm = ChatOpenAI(
                model="gpt-4o",
                temperature=0,
                max_tokens=None,
                timeout=None,
                max_retries=2
            )
        
    if llm is not None:
        message = f"Be precise in your response. Here is the question: {message}"
        response = llm.invoke(message)
    else:
        return "No LLM chosen"
    
    return response.content

def format_message(message, role):
    """
    """
    if role=='user':
        return (
            'human', message
        )
    else:
        return (
            'assistant', message
        )


def show_spinner(stop_event):
    """Display a rotating cursor."""
    spinner = itertools.cycle(["|", "/", "-", "\\"])
    while not stop_event.is_set():
        sys.stdout.write(f"\rExecuting... {next(spinner)}")
        sys.stdout.flush()
        time.sleep(0.1)
    sys.stdout.write("\rExecution complete.      \n")
    sys.stdout.flush()


def execute_command(command, message_printer=None):
    if message_printer:
        stop_event = threading.Event()
        spinner_thread = threading.Thread(target=show_spinner, args=(stop_event,))
        message_printer.system_message(f"\nExecuting command -  {command}\n")
        spinner_thread.start()
    result = subprocess.run(command, shell=True, capture_output=True, text=True)

    if message_printer:
        stop_event.set()
        spinner_thread.join()
    return result


def get_response(response, message_printer, session):

    try:

        message_printer.assistant_message(response)

        commands = extract_command(response)

        if commands:
            for i, command in enumerate(commands, start=1):
                res = confirm_ask(i)
                if res:
                    if res=='edit':
                        text = HTML(f'<ansiblue>> {command}</ansiblue> ')

                        command = session.prompt("\nedit: ", default=command, placeholder='')

                    result = execute_command(command, message_printer)
                    if result.stderr:
                        print("Error while executing")
                        message_printer.assistant_message(result.stderr, type='command')
                    else:
                        # print_response(f"command executed \n {result.stdout}", type='command')
                        message_printer.assistant_message(f"{result.stdout}", type='command')

            # # ask
            
            
            #     if res=='edit':
                    
                
            #         stop_event = threading.Event()
            #         spinner_thread = threading.Thread(target=show_spinner, args=(stop_event,))
            #         # print_response(command)
            #         message_printer.system_message(f"\nExecuting command -  {command}\n")
            #         spinner_thread.start()
            #         result = subprocess.run(command, shell=True, capture_output=True, text=True)
            #         # Stop the spinner
            #         stop_event.set()
            #         spinner_thread.join()
                    
            #         if result.stderr:
            #             print("Error while executing")
            #             message_printer.assistant_message(result.stderr, type='command')
            #         else:
            #             # print_response(f"command executed \n {result.stdout}", type='command')
            #             message_printer.assistant_message(f"{result.stdout}", type='command')

    except Exception as e:
        log.error("Error:", e)

def confirm_ask(command_index):
    """
    Prompt the user to decide whether to run or not.
    """
    if command_index == 1:
        command_txt = f"command"
    elif command_index == 2:
        command_txt = f"2nd command"
    elif command_index == 3:
        command_txt = f"3rd command"
    else:
        command_txt = f"{command_index}th command"

    ask_text = f"\n[bold cyan]Do you want to run {command_txt}?[/bold cyan] [green](Yes (y) / No (n))[/green] or [yellow]Edit (e):[/yellow]"
    # console.print(ask_text, end=" ")
    response = console.input(ask_text).strip().lower()
    if response in ['yes', 'y']:
        return True
    elif response in ['no', 'n']:
        return False
    elif response in ['edit', 'e']:
        return "edit"
    else:
        print("Invalid input. Please respond with 'Yes' or 'No'.")
        confirm_ask(command_index)  # Re-prompt the user

def extract_command(text):
    """
    Extracts all commands enclosed within ```bash ... ``` from the given text.

    Args:
        text (str): The input string containing commands.

    Returns:
        list: A list of extracted commands as strings.
    """
    # Regular expression to match text inside triple backticks with bash
    command_pattern = r"```bash\s+([\s\S]*?)```"
    commands = re.findall(command_pattern, text)
    # Strip leading and trailing whitespace from each command
    return [cmd.strip() for cmd in commands]

# Example usage
if __name__ == "__main__":
    printer = MessagePrinter()

    # Print system messages
    printer.print_system_message("Type commands as usual. Ask anything you want.")
    printer.print_system_message("Type 'exit' or 'quit' to stop the program.\n")

    # Print user message
    printer.print_user_message("Hello, how do I use this program?")

    # Print assistant message
    printer.print_assistant_message("Sure! Let me explain how it works.")

    # Print messages with Markdown rendering
    printer.print_user_message("### This is a Markdown heading")
    printer.print_assistant_message("**This text is bold in Markdown**")




from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.shortcuts import CompleteStyle, PromptSession
from prompt_toolkit.styles import Style

import os

PLACEHOLDER = [('gray', 'ask me anything')]

FILE_DISPLAY_PLACEHOLDER = [('gray', '')]


class FilePathCompleter(Completer):

    def handle_tags(self, text):
        """
        """
        if '!' in text:
            if len(text)==1:
                text = ''
            else:
                text = text[1:]
        return text

    def get_completions(self, document, complete_event):
        text = document.text_before_cursor

        text = self.handle_tags(text)

        word = text.split(" ")[-1]

        dirname, partial_name = os.path.split(word)

        if not dirname:  # Default to the current directory if none is specified
            dirname = "."
        
        try:
            # List entries in the directory
            if os.path.isdir(dirname):
                entries = os.listdir(dirname)
            else:
                entries = []
        except FileNotFoundError:
            return

        # Filter entries based on partial matching
        for entry in entries:
            full_path = os.path.join(dirname, entry)
            if partial_name.lower() in entry.lower():  # Case-insensitive substring match
                # Add a trailing '/' to directories
                display = entry if os.path.isdir(full_path) else entry                
                yield Completion(
                    display,
                    start_position=-len(partial_name),
                    # display_meta="Directory" if os.path.isdir(full_path) else "File",
                )


# Custom key bindings to handle TAB key press
kb = KeyBindings()


@kb.add(Keys.Tab)
def _(event):
    "Handle tab key press for completion."
    b = event.current_buffer
    if b.complete_state:
        b.complete_next()
    return True


custom_style = Style.from_dict({
    'prompt': 'bold green',  # Style for the prompt text
    '': 'white',           # Default text style
})


session = PromptSession(completer=FilePathCompleter(), 
                        key_bindings=kb, 
                        complete_style=CompleteStyle.MULTI_COLUMN,
                        style=custom_style
                        )



def main():


    session = PromptSession(completer=FilePathCompleter(), 
                            key_bindings=kb, 
                            complete_style=CompleteStyle.MULTI_COLUMN,
                            style=custom_style
                            )
    while True:
        try:
            user_input = session.prompt("Enter path: ")
            print(f"You selected: {user_input}")
        except (KeyboardInterrupt, EOFError):
            break


if __name__ == "__main__":
    main()




import uvicorn
from threading import Thread
import asyncio
from fastapi import FastAPI
import time
from queue import Queue



class FastAPIServer:
    def __init__(self, app: FastAPI, host: str = "127.0.0.1", port: int = 8000):
        self.app = app
        self.host = host
        self.port = port
        self.server = None
        self.thread = None
        self.loop = None
        self.data_queue = Queue()
        self.running = False

    def start(self):
        """Starts the FastAPI server in a separate thread."""

        self.running = True

        if self.thread and self.thread.is_alive():
            # print("FastAPI server is already running.")
            return False

        def run_server():
            """Run the server inside an event loop."""
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
            config = uvicorn.Config(self.app, host=self.host, port=self.port, log_level="error")
            self.server = uvicorn.Server(config)
            self.loop.run_until_complete(self.server.serve())

        self.thread = Thread(target=run_server, daemon=True)
        self.thread.start()
        # print(f"FastAPI server started on http://{self.host}:{self.port}")

    def stop(self):
        """Stops the FastAPI server."""
        self.running = False
        if self.server and self.loop:

            # print("Stopping FastAPI server...") # needs to be added to log file
            self.loop.call_soon_threadsafe(lambda: setattr(self.server, "should_exit", True))
            self.thread.join()  # Wait for the thread to finish

            if self.thread.is_alive():
                # print("thread is still alive")
                pass

    def send_data(self, data):
        """Send data to the queue."""
        self.data_queue.put(data)





import os
from pydantic import BaseModel, Field
from typing import List
from .dependencies import BAM_VISUALIZATION

def plan(input_filepath, session, message_printer):
    filename, file_ext = os.path.splitext(input_filepath)
    if file_ext == '.bam':
        if os.path.exists(input_filepath):
            file_requirements = bamFileReq(input_filepath, file_ext, session, message_printer)
            config = generate_config(file_ext, file_requirements)
            return config
        return None

def generate_config(ext, file_requirements):
    """
    """
    if ext=='.bam':
        assembly = {
                "name": 'NC_045512',

                "sequence": {
                    "type": 'ReferenceSequenceTrack',
                    "trackId": 'GRCh38-ReferenceSequenceTrack',
                    "adapter": {
                        "type": 'IndexedFastaAdapter',
                        "fastaLocation": {
                            "uri": f'http://127.0.0.1:8000/uploads/{file_requirements[".fa"]}',
                            },
                        "faiLocation": {
                            "uri": f'http://127.0.0.1:8000/uploads/{file_requirements[".fai"]}',
                            },
                        },
                    },
                }
        track = [{
        "type": 'AlignmentsTrack',
        "trackId": "genes", 
        "name": 'spike-in_bams_file_0.bam',
        "assemblyNames": ['NC_045512'],
        "category": ['Genes'],
        "adapter": {
          "type": 'BamAdapter',
          "bamLocation": {
            "uri": f'http://127.0.0.1:8000/uploads/{file_requirements[".bam"]}',
          },
          "index": {
            "location": {
              "uri": f'http://127.0.0.1:8000/uploads/{file_requirements[".bai"]}',
            },
          },
        }
    }]
    
    return {
        "assembly": assembly, 
        "track": track
        }



def bamFileReq(bamfile, ext, session, message_printer):
    """
    """
    requirements = BAM_VISUALIZATION

    prefix_text = f"In order display {bamfile}. \nFollowing input file types are required: \n "
    prefix_text += "\n".join([f"{i}. {key} file" for i, (key, value) in enumerate(requirements.items()) if key != ext])

    message_printer.assistant_message(prefix_text)



    all_file_quries = []

    
    for i, (key, value) in enumerate(requirements.items()):
        if key == ext:
            input_file_path = bamfile
        else:
            input_file_path = ""
        
        file_query = FileQuery(
            id=i,
            file_path=input_file_path,  # Replace with an actual file path
            file_type=key
        )
        all_file_quries.append(file_query)

    visualization_planner = VisualizationPlanner(file_queries=all_file_quries)
    
    all_files_required = visualization_planner.execute(session, message_printer)
    return all_files_required

class FileQuery(BaseModel):
    """
    """

    id: int = Field(..., description="Unique id of the query")
    file_path: str = Field(default=None, description="file required to display")
    file_type: str = Field(..., description="file extenstion")

    dependancies: int = Field(
        default=0,
        description="not needed now but maybe in future")
    
    def check_file(self, input_file):
        if input_file:
            filename, file_ext = os.path.splitext(input_file)
            if os.path.exists(input_file) and file_ext == self.file_type:
                return True
        return False

    def execute(self, required_files, session, message_printer):
        """
        """ 
        if self.check_file(self.file_path):
            required_files[self.file_type] = self.file_path
            return required_files
        else:
            while True:
                user_input = session.prompt(f"\n Please provide the {self.file_type} file : ")
                
                if self.check_file(user_input):
                    required_files[self.file_type] = user_input
                    return required_files
                else:
                    message_printer.assistant_message("Please provide the correct file.")

class VisualizationPlanner(BaseModel):
    """
    """
    file_queries: List[FileQuery] = Field(
        ..., description="list of files required to get from user"
    )
    required_files: dict = Field(default={}, description="required files")

    def dependencies(self, id: List[int]) -> List[FileQuery]:
        """
        not needed for now. but might be neded in future. 
        """
        pass

    def execute(self, session, message_printer):
        """
        """
        for query in self.file_queries:
            res = query.execute(
                self.required_files, session, message_printer
            )
            self.required_files = res
        return self.required_files
    
    
                


import cProfile
import os
import subprocess
import shlex
import argparse
import re
import random
import subprocess

from nucleus.data_viz.planner import plan
from nucleus.terminal.suggestion import session, PLACEHOLDER
from nucleus.terminal.planner import QueryManager
from nucleus.tools import MessagePrinter, execute_command
from nucleus.logger import log

# data viz server
from nucleus.data_viz.server import FastAPIServer
from nucleus.data_viz.app import app


# data_viz_server = FastAPIServer(app)
# app.state.server_instance = data_viz_server

# data_viz_server.start()

file_requirements = {
    '.bam': "",
    '.bai': "",
    '.fa': "",
    '.fai': ""
}

def args_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--openai-api-key",  help="Specify the OpenAI API key")
    parser.add_argument("--anthropic-api-key", help="Specify the Anthropic API key")
    args = parser.parse_args()

    model, api_key = None, None

    input_vals = {
        'LLM': []
    }

    if args.openai_api_key:
        model = 'openai'
        api_key = args.openai_api_key

        input_vals['LLM'] = [{
            'model': model, 
            "api_key": api_key
            }]
        
        os.environ["OPENAI_API_KEY"] = api_key
        
        return input_vals
    
    if args.anthropic_api_key:
        model = "anthropic"
        api_key = args.anthropic_api_key

        os.environ["ANTHROPIC_API_KEY"] = api_key

        input_vals['LLM'] = [{
            'model': model, 
            "api_key": api_key
            }]
        
        return input_vals

    openai_api_key = os.getenv("OPENAI_API_KEY")
    anthropic_api_key = os.getenv("ANTHROPIC_API_KEY")
    gemini_api_key = os.getenv("GEMINI_API_KEY")
    
    

    if openai_api_key:
        model='openai'
        api_key = openai_api_key
        
        input_vals['LLM'].append({
            "model":model,
            "api_key": api_key
        })
    
    if anthropic_api_key:
        model='anthropic'
        api_key = anthropic_api_key

        input_vals['LLM'].append({
            "model":model,
            "api_key": api_key
        })
    
    if gemini_api_key:
        model='gemini'
        api_key = gemini_api_key

        input_vals['LLM'].append({
            "model":model,
            "api_key": api_key
        })
    
    return input_vals
    

def main():
    """
    """

    interrupted_once = False
    input_vals = {}

    input_args = args_parser()
    
    if not len(input_args['LLM']):
        log.error("Error: No API key provided. Please provide an API key to proceed.")
        return

    if len(input_args['LLM'])==1:
        input_vals['LLM'] = input_args['LLM'][0]

    if len(input_args["LLM"])>1:
        log.warning("Multiple API keys found in your environment variables, but none explicitly provided.")
        llm = random.choice(input_args["LLM"])
        log.info(f"Randomly selected API key for the model '{llm['model']}'.")
        input_vals['LLM'] = llm

    message_printer = MessagePrinter()

    message_printer.system_message("\nExecute terminal commands with prefix '!'. Ask questions as usual.")
    message_printer.system_message("Type 'exit' or 'quit' to terminate the program. \n")


    query_responder = QueryManager(input_vals, message_printer, session)
    while True:
        try:
            # Prompt for user input

            user_input = session.prompt("\n> ", placeholder=PLACEHOLDER)

            if user_input:
                interrupted_once = False

            # Check if the user wants to exit
            if user_input.lower() == "":
                continue

            if user_input.lower() in ["exit", "quit"]:
                # data_viz_server.stop()
                message_printer.system_message("Exiting shell. Goodbye!")
                break

            if user_input.lower()[0] in "!":
                # handle 'cd' command to change directory
                user_input = user_input[1:]
                if user_input.startswith("cd"):
                    parts = shlex.split(user_input)
                    if len(parts) > 1:
                        new_dir = parts[1]
                    else:
                        new_dir = os.path.expanduser("~")
                    try:
                        os.chdir(new_dir)
                    except FileNotFoundError:
                        print(f"cd: {new_dir}: No such file or directory")
                    except Exception as e:
                        print(f"cd: {e}")
                else:
                    # subprocess.run(shlex.split(user_input), check=False)
                    result = execute_command(user_input)

                    # result = subprocess.run(user_input, shell=True, capture_output=True, text=True)
                    if result.stderr:
                        print("Error while executing")
                        message_printer.assistant_message(result.stderr, type='command')
                    else:
                        # print_response(f"command executed \n {result.stdout}", type='command')
                        message_printer.assistant_message(f"{result.stdout}", type='command')
                        

            # elif user_input.startswith("/show"):
            #     file_input = user_input.split("/show", 1)[-1].strip()
            #     config = plan(file_input, session, message_printer)
            #     if config:
            #         # data_queue.put(config)
            #         data_viz_server.send_data(config)

            #         message_printer.system_message("You can view the file at http://localhost:8000")
            #     else:
            #         message_printer.system_message("\n Provided file doesn't exist.")
            #         continue
            
            # elif user_input == "/close":
            #     data_viz_server.stop()
            else:
                query_responder.execute_query(user_input)

        except KeyboardInterrupt:
            # try:
            if interrupted_once:
                # data_viz_server.stop()
                message_printer.system_message("\n Exiting shell. Goodbye!")
                break
            else:
                interrupted_once = True
                message_printer.system_message("\n ^C again to exit")
                continue
        except Exception as e:
            log.error(f"Error : {e}")


if __name__ == "__main__":
    main()


import os
from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from fastapi.staticfiles import StaticFiles
from fastapi.responses import FileResponse
from fastapi.middleware.cors import CORSMiddleware
import asyncio
from pathlib import Path


app = FastAPI()

server_instance = None

# Add CORS middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Allows all origins
    allow_credentials=True,
    allow_methods=["*"],  # Allows all methods
    allow_headers=["*"],  # Allows all headers
)

react_build_dir = Path(__file__).parent / "../visualization/dist"

uploads_dir = os.curdir

app.mount("/uploads", StaticFiles(directory=uploads_dir), name="uploads")

app.mount("/assets", StaticFiles(directory=react_build_dir / "assets"), name="assets")

# Serve the index.html for the root endpoint
@app.get("/")
async def serve_react_app():
    index_file = react_build_dir / "index.html"
    return FileResponse(index_file)


@app.get("/status")
async def status():
    return {"status": "Running"}

@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    try:
        server_instance = websocket.app.state.server_instance
        while True:
            if server_instance and not server_instance.data_queue.empty():
                data_to_send = server_instance.data_queue.get()
                try:
                    await asyncio.sleep(1)
                    await websocket.send_json(data_to_send)
                    # server_instance.data_queue.put(data_to_send)

                except WebSocketDisconnect:
                    print("WebSocket connection closed while sending data.")
                    # break
                except Exception as e:
                    print(f"Error while sending data: {e}")
                    break
            else:
                # Check for a condition to break the loop to avoid infinite wait
                if not server_instance.running:
                    print("Exiting loop due to condition.")
                    break
                await asyncio.sleep(0.1)  # Allow the loop to wait briefly
                
    except WebSocketDisconnect:
        print("WebSocket connection closed by the client.")
    except Exception as e:
        print(f"Unexpected error: {e}")
    finally:
        print("WebSocket endpoint cleanup.")




import sys

def model_prompt():
    return (
        "system",
        """
        You are a helpful assistant specialized in generating bioinformatics command-line instructions. Your task is to assist the user by providing accurate and efficient terminal commands in response to their queries. Follow these guidelines:

        1. If the user asks for a specific bioinformatics task, respond with the corresponding command enclosed in a bash markdown block. For example:
        - USER: Convert haplotype.sam to haplotype.bam
            ASSISTANT:
            ```bash
            samtools view -S -b haplotype.sam > haplotype.bam
            ```
        - USER: List files in the current directory
            ASSISTANT:
            ```bash
            ls
            ```

        2. If the query is unclear or you do not understand the user's request, politely ask for clarification.

        Ensure your responses are precise and formatted correctly for terminal usage.
        """
        )

def planner_prompt():
    """
    """
    return [(
                "system",
                """
                You are an expert query planner capable of breaking down complex questions into smaller, dependent subqueries to create a clear path for answering the main question. Your role involves:  

                    1. Identifying the type of query:  
                    - **Command-related queries:** Provide terminal commands to accomplish specific tasks using bioinformatics tools. Users may specify file paths for input or output files in their queries.
                    - **General bioinformatics/biology-related questions:** Create a structured plan to answer them effectively.  

                    2. Generating a clear plan of subqueries:  
                    - Break the problem into its dependencies, listing subqueries in order from lowest dependency to highest.  
                    - For multiple tasks in the same query, treat them as sequential steps and plan accordingly.  

                    3. Rules for responses:  
                    - Do **not** answer the question directly; instead, provide a step-by-step query plan to address the problem.  
                    - Before providing the plan, carefully think through the problem to ensure a thorough and accurate breakdown.  
                    - If the query is unclear, ask the user for clarification.  
                    - Always design concise and manageable plans for clarity and efficiency.  

                    By adhering to these guidelines, you will help users tackle complex tasks systematically and execute terminal commands effectively.

                """,
            ),
            (
                "user",
                """
                Consider the following question: {question}.  
                    Generate an appropriate query plan based on the dependencies involved.  
                    - If the query does not rely on any other subqueries, classify it as a **SINGLE_QUESTION** query type.  
                    - If the query requires multiple subqueries or has dependencies, classify it as a **MULTI_DEPENDENCY** query type.  
                """
            )
        ]



from rich.console import Console
from rich.logging import RichHandler
from pathlib import Path
import logging

# Setup Rich Console
console = Console()

# Setup Logging
LOG_FILE = "/.nucleus/nucleus.log"
logging.basicConfig(
    level=logging.ERROR,  # Set the default log level
    format="%(message)s",
    handlers=[
        RichHandler(console=console, rich_tracebacks=True, markup=True),  # Rich formatting for console
        # logging.FileHandler(LOG_FILE)  # Save logs to a file
    ]
)

# Get Logger
log = logging.getLogger("rich_logger")

# Example log messages
def log_examples():
    log.debug("This is a [bold green]DEBUG[/bold green] message for internal debugging.")
    log.info("This is an [cyan]INFO[/cyan] message for general information.")
    log.warning("This is a [yellow]WARNING[/yellow] message indicating potential issues.")
    log.error("This is an [bold red]ERROR[/bold red] message for serious problems.")
    log.critical("This is a [bold red on white]CRITICAL[/bold red on white] message for severe issues.")

if __name__ == "__main__":

    log_examples()
    console.print(f"[green]Log messages have been saved to[/green] {LOG_FILE}.")



import os
from pydantic import BaseModel, Field
from typing import List
import enum
from langchain_openai import ChatOpenAI
from nucleus.tools import get_response, format_message
from langchain_core.prompts import ChatPromptTemplate
from nucleus.logger import log
from nucleus.tools import command_provider
from nucleus.prompts import model_prompt, planner_prompt

class ComputeQuery(BaseModel):
    """
    computation of a query
    """
    query: str
    response: dict = {}
    done: bool = False

    def execute(self, input_dict):
        """
        """
        message = []
        message.append(model_prompt())
        message.append(format_message(self.query, role='user'))

        model_name = input_dict["LLM"]['model']
        try:
            result = command_provider(self.query, model_name)
            self.response = result
            if self.response:
                self.done = True
        except Exception as e:
            log.error("Exception in compute query", e)


class MergedResponses(BaseModel):
    """
    Models a merged response of multiple queries.
    Currently we just concatinate them but we can do much more complex things.
    """
    dependent_queries: list[ComputeQuery]
    query: str = '...'

    def execute(self, input_dict):
        """
        """
        prompt = ""
        for dependent in self.dependent_queries:
            prompt += f"Assistant: {dependent.query} \n"
            prompt += f"User: {dependent.response} \n"


        # prompt = ""
        # "\n".join([f"Question: {subquries.query} \n Answer: {subquries.response}"  for subquries in self.dependent_queries])
        
        self.query = f"These are the list of questions and its' responses. {prompt} Based on this, answer this Question: {self.query}"

        merger_query = ComputeQuery(query=self.query)
        merger_query.execute(input_dict)

        return merger_query

class QueryType(str, enum.Enum):
    """
    Enumeration representing the types of queries that can be asked to a question answer system.
    """
    # When i call it anything beyond 'merge multiple responses' the accuracy drops significantly.
    SINGLE_QUESTION = "SINGLE"
    MULTI_DEPENDENCY = "MULTI_DEPENDENCY"

class QueryTask(BaseModel):
    """
    """
    id: int = Field(..., description="unique id to the query")
    query: str = Field(default=None, description="the sub query")
    dependencies: list[int] = Field(
        default = [],
        description="no of dependecies to respond to this query"
    )
    query_type: QueryType = Field(
        default=QueryType.SINGLE_QUESTION,
        description="Type of question, either a single question or a multi question merge when there are multiple questions",
    )
    
    answer : str = ""
    completed: str = False

    def execute(self, input_dict, dependency_func):

        if self.completed:
            return self.answer
    
        if self.query_type == QueryType.SINGLE_QUESTION:
            compute_query = ComputeQuery(query=self.query,)
            compute_query.execute(input_dict)
            self.completed = True

            self.answer = compute_query

            return self.answer
        else:
            sub_queries = dependency_func(self.dependencies)
            computed_queries = [q.execute(input_dict, dependency_func) for q in sub_queries]

            merge_responses = MergedResponses(query=self.query, dependent_queries=computed_queries)
            
            self.answer = merge_responses.execute(input_dict)
            self.completed = True
            return self.answer

class QueryPlanner(BaseModel):
    """
    """
    query_tasks: List[QueryTask] = Field(..., description="List of subqueries to the main from the user.")

    def dependencies(self, idz: list[int]) -> list[QueryTask]:
        """
        Returns the dependencies of the query with the given id.
        """
        return [q for q in self.query_tasks if q.id in idz]

    def execute(self, input_dict):
        """
        """
        result = []

        for query in self.query_tasks[::-1]:

            if not query.completed:
                query.execute(input_dict,
                    dependency_func=self.dependencies,
                )
                if query.completed:
                    result.append(query.answer)

        return result
     
class QueryManager:
    """
    """
    planner_prompt = ChatPromptTemplate.from_messages(
        planner_prompt()
        )
    

    def __init__(self, input_dict, message_printer, session):
        """
        """

        self. planner = self.planner_prompt | ChatOpenAI(
            model="gpt-4o", temperature=0,
            ).with_structured_output(QueryPlanner)

        self.input_dict = input_dict
        self.model = input_dict['LLM']['model']
        self.message_printer = message_printer
        self.session = session

    def handle_response(self, results):
        """
        """

        if len(results)>1:

            responses = "\n".join([res.response for res in results])
            prompt = "rewrite these: "
            prompt += responses
            prompt += "for this question: \n"
            prompt += self.input_dict['question']
            merged_response = command_provider(prompt, self.model)
         
        else:
            merged_response = results[0].response

        get_response(merged_response, self.message_printer, self.session)

    def execute_query(self, question):
        """
        """
        try:

            self.input_dict['question'] = question
            plan = self.planner.invoke({"question": question})
            result = plan.execute(self.input_dict)
            
            self.handle_response(result)

        except Exception as e:
            print("error", e)



if __name__ == '__main__':
    while True:
        user_input = input("> ")
        input_dict = {}

        input_dict['LLM'] = {'model':'openai'}
        input_dict['question'] = user_input
        plan = QueryManager(input_dict)
        print()
        plan.execute_query(user_input)




from langchain_openai import ChatOpenAI
from langchain_anthropic import ChatAnthropic



def load_model(input_dict):
    """
    """

    model_name = input_dict['model']
    if model_name == 'openai':
        llm = ChatOpenAI(
            model="gpt-4o",
            temperature=0,
            max_tokens=None,
            timeout=None,
            max_retries=2,
            api_key=input_dict['api_key']
        )

    if model_name == 'anthropic':
        llm = ChatAnthropic(
            model="claude-3-opus-20240229",
            temperature=0,
            api_key=input_dict['api_key']
        )

    return llm